#~ '(C) Links System Software 2009
#~ 'Feedback information
#~ 'www: http://lss2008.livejournal.com/
#~ 'E-mail1: designer@ls-software.ru
#~ 'E-mail2: kirill2007_77@mail.ru (search this e-mail to contact via skype)
#~ 'icq: 328-958-369

#~ mshstick.rb ver. 1.2. 02-Jan-10
#~ Plug-in, which deforms selected group of faces to follow the selected surface

#~ How to use:
#~ 1)	put mshstick.rb to Google SketchUp 6\Plugins folder
#~ 2)	start SketchUp 6
#~ 3) select some groups of faces, which you want to stick to another group of faces (surface mesh)
#~ 4) choose the command "Plugins>>Stick selected groups to mesh group" command from the main menu
#~ 5) pick the group with mesh, to which you are going to stick selected groups
#~ 6) specify sticking options within appeared dialog and click 'OK' to start sticking
#~ 7) if 'Specify manually' options for 'Stick direction' and 'Reflection ray direction' were chosen, then specify 'Stick direction' first
#~     and 'Reflection ray direction' after

require 'sketchup.rb'

class Lss_Specify_Dir_Vec_Tool
    attr_accessor :dir_vec
    def initialize(prev_tool,vec_name)
      @prev_tool = prev_tool
      @vec_name=vec_name
      @ip1 = nil
      @ip2 = nil
      @xdown = 0
      @ydown = 0
    end
    
    def activate
      @ip1 = Sketchup::InputPoint.new
      @ip2 = Sketchup::InputPoint.new
      @ip = Sketchup::InputPoint.new
      @drawn = false
      self.reset(nil)
    end

    def deactivate(view)
      view.invalidate if @drawn
    end
    
    def onMouseMove(flags, x, y, view)
      if( @state == 0 )
        @ip.pick view, x, y
        if( @ip != @ip1 )
          view.invalidate if( @ip.display? or @ip1.display? )
          @ip1.copy! @ip
          view.tooltip = @ip1.tooltip
        end
      else
        @ip2.pick view, x, y, @ip1
        view.tooltip = @ip2.tooltip if( @ip2.valid? )
        view.invalidate
        if( (x-@xdown).abs > 10 || (y-@ydown).abs > 10 )
          @dragging = true
        end
      end
    end
    
    def onLButtonDown(flags, x, y, view)
      if( @state == 0 )
        @ip1.pick view, x, y
        if( @ip1.valid? )
          @state = 1
          Sketchup::set_status_text $exStrings.GetString("Select second end"), SB_PROMPT
          @xdown = x
          @ydown = y
        end
      else
        if( @ip2.valid? )
          self.create_geometry(@ip1.position, @ip2.position,view)
          self.reset(view)
        end
      end
      view.lock_inference
    end
    
    def onKeyDown(key, repeat, flags, view)
      if( key == CONSTRAIN_MODIFIER_KEY && repeat == 1 )
        @shift_down_time = Time.now
        if( view.inference_locked? )
          view.lock_inference
        elsif( @state == 0 && @ip1.valid? )
          view.lock_inference @ip1
        elsif( @state == 1 && @ip2.valid? )
          view.lock_inference @ip2, @ip1
        end
      end
    end
  
  def onKeyUp(key, repeat, flags, view)
    if( key == CONSTRAIN_MODIFIER_KEY &&
      view.inference_locked? &&
      (Time.now - @shift_down_time) > 0.5 )
      view.lock_inference
    end
  end
  
    def onLButtonUp(flags, x, y, view)
      if( @dragging && @ip2.valid? )
        self.create_geometry(@ip1.position, @ip2.position,view)
        self.reset(view)
      end
    end
    
    def draw(view)
      if( @ip1.valid? )
        if( @ip1.display? )
          @ip1.draw(view)
          @drawn = true
        end
        
        if( @ip2.valid? )
          @ip2.draw(view) if( @ip2.display? )
          view.set_color_from_line(@ip1, @ip2)
          self.draw_geometry(@ip1.position, @ip2.position, view)
          @drawn = true
        end
      end
    end
    
    def reset(view)
      @state = 0
      Sketchup::set_status_text($exStrings.GetString("Select first end"), SB_PROMPT)
      @ip1.clear
      @ip2.clear
      if( view )
        view.tooltip = nil
        view.invalidate if @drawn
      end
      @drawn = false
      @dragging = false
    end
  
    def onCancel(flag, view)
      self.reset(view)
    end
    
    def draw_geometry(pt1, pt2, view)
      view.draw_line(pt1, pt2)
    end
  
    def create_geometry(p1, p2, view)
      @dir_vec=p1.vector_to(p2)
      @prev_tool.stick_ray_vec=@dir_vec.normalize if @vec_name=="stick ray"
      @prev_tool.refl_ray_vec=@dir_vec.normalize if @vec_name=="reflection ray"
      Sketchup.active_model.select_tool @prev_tool
    end
end #class Lss_Specify_Dir_Vec_Tool

class Lss_Msh_Stick_Tool
    attr_accessor :stick_ray_vec
    attr_accessor :mshstick_tool
    attr_accessor :refl_ray_vec
    
    def initialize
      @entities=Sketchup.active_model.active_entities
      @selection=Sketchup.active_model.selection
      @input_results=nil
      @refl_ray_vec=nil
    end
    
    def activate
      if @input_results
        if @refl_ray_opt=="Specify manually" and @refl_ray_vec.nil?
          Sketchup.set_status_text("Draw line to specify reflection ray direction.")
          specify_refl_vec=Lss_Specify_Dir_Vec_Tool.new(@mshstick_tool,"reflection ray")
          Sketchup.active_model.select_tool specify_refl_vec
          puts "Specify reflection vector direction."
        else
          puts "Stick process started."
          add_edges_tool=Lss_Add_Edges_Tool.new(@msh_group,@grps_set,@stick_ray_vec,@check_self_int,@stick_type,@refl_ray_vec,@offset_val,@magnify_val)
          Sketchup.active_model.select_tool add_edges_tool
          view=Sketchup.active_model.active_view
          if @skip_edg_proj=="No"
            add_edges_tool.add_proj_edges(view)
          else
            add_edges_tool.skip_proj_edges(view)
          end
          @input_results=nil
        end
      else
        @grps_set=Set.new
        @selection.each{|ent|
          if ent.typename=="Group"
            @grps_set.insert ent
          end
        }
        @selection.clear
        @ip = Sketchup::InputPoint.new
        @ip1 = Sketchup::InputPoint.new
      end
    end
    
    def onMouseMove(flags, x, y, view)
      @ip1.pick view, x, y
      if( @ip1 != @ip )
          view.invalidate
          @ip.copy! @ip1
          view.tooltip = @ip.tooltip
      end
      ph=view.pick_helper
      ph.do_pick x,y
      @b_p = ph.best_picked
      if @b_p
        Sketchup.set_status_text("#{@b_p.typename} is under cursor.")
      else
        Sketchup.set_status_text("Click on group, which contains surface.")
      end
    end
    
    def draw(view)
      if @b_p
        if @b_p.typename=="Group"
          self.draw_grp_bnds(@b_p.bounds,view)
        else
          txt_pt3d=Geom::Point3d.new(@b_p.bounds.center)
          txt_pt = view.screen_coords(txt_pt3d)
          view.draw_text(txt_pt, "This is '#{@b_p.typename}'")
          self.draw_no_grp_bnds(@b_p.bounds,view)
        end
      end
    end
    
    def draw_grp_bnds(grp_bounds,view)
      corner_no=0
      while corner_no<=7
        view.draw_points grp_bounds.corner(corner_no), 6, 1,"green"
        corner_no+=1
      end
    end
    
    def draw_no_grp_bnds(bnds,view)
      corner_no=0
      while corner_no<=7
        view.draw_points bnds.corner(corner_no), 6, 4,"red"
        corner_no+=1
      end
    end 
    
    def onLButtonUp(flags, x, y, view)
      @refl_ray_vec=nil
      ph=view.pick_helper
      ph.do_pick x,y
      @b_p = ph.best_picked
      if not @b_p.nil?
        puts "#{@b_p.typename} is picked."
        Sketchup.set_status_text("#{@b_p.typename} is picked. Click on group with surface (green square bound marks).")
        if @b_p.typename == "Group"
          @msh_group=@b_p
          prompts = ["Perform checking for non-terrain surface ", "Stick direction ", "Stick type ", "Skip edges projection stage ", "Reflection ray direction ", "Offset value ", "Magnify "]
          defaults= ["No","Down","Normal","No","Parallel to stick ray",0.0,1.0]
          list = ["Yes|No","Down|Up|Left|Right|Front|Back|Specify manually","Normal|Super","No|Yes","Parallel to stick ray|Normally to surface|Specify manually","",""]
          @input_results=UI.inputbox prompts, defaults, list, "Groups Sticking Settings"
          if @input_results==false
            puts "Sticking was cancelled."
            Sketchup.active_model.select_tool nil
            return
          end
          @check_self_int,@stick_dir,@stick_type,@skip_edg_proj,@refl_ray_opt,@offset_val,@magnify_val=@input_results
          case @refl_ray_opt
            when "Parallel to stick ray"
            @refl_ray_vec=Geom::Vector3d.new(0,0,0)
            when "Normally to surface"
            @refl_ray_vec=Geom::Vector3d.new(1,1,1)
          end
          case @stick_dir
            when "Down"
            Sketchup.set_status_text("#{@b_p.typename} is picked.")
            @stick_ray_vec=Geom::Vector3d.new(0,0,-1)
            if @refl_ray_opt=="Specify manually"
              Sketchup.set_status_text("Draw line to specify reflection ray direction.")
              specify_refl_vec=Lss_Specify_Dir_Vec_Tool.new(@mshstick_tool,"reflection ray")
              Sketchup.active_model.select_tool specify_refl_vec
            else
              add_edges_tool=Lss_Add_Edges_Tool.new(@msh_group,@grps_set,@stick_ray_vec,@check_self_int,@stick_type,@refl_ray_vec,@offset_val,@magnify_val)
              Sketchup.active_model.select_tool add_edges_tool
              if @skip_edg_proj=="No"
                add_edges_tool.add_proj_edges(view)
              else
                add_edges_tool.skip_proj_edges(view)
              end
            end
            when "Up"
            Sketchup.set_status_text("#{@b_p.typename} is picked.")
            @stick_ray_vec=Geom::Vector3d.new(0,0,1)            
            if @refl_ray_opt=="Specify manually"
              Sketchup.set_status_text("Draw line to specify reflection ray direction.")
              specify_refl_vec=Lss_Specify_Dir_Vec_Tool.new(@mshstick_tool,"reflection ray")
              Sketchup.active_model.select_tool specify_refl_vec
            else
              add_edges_tool=Lss_Add_Edges_Tool.new(@msh_group,@grps_set,@stick_ray_vec,@check_self_int,@stick_type,@refl_ray_vec,@offset_val,@magnify_val)
              Sketchup.active_model.select_tool add_edges_tool
              if @skip_edg_proj=="No"
                add_edges_tool.add_proj_edges(view)
              else
                add_edges_tool.skip_proj_edges(view)
              end
            end
            when "Left"
            Sketchup.set_status_text("#{@b_p.typename} is picked.")
            @stick_ray_vec=Geom::Vector3d.new(-1,0,0)            
            if @refl_ray_opt=="Specify manually"
              Sketchup.set_status_text("Draw line to specify reflection ray direction.")
              specify_refl_vec=Lss_Specify_Dir_Vec_Tool.new(@mshstick_tool,"reflection ray")
              Sketchup.active_model.select_tool specify_refl_vec
            else
              add_edges_tool=Lss_Add_Edges_Tool.new(@msh_group,@grps_set,@stick_ray_vec,@check_self_int,@stick_type,@refl_ray_vec,@offset_val,@magnify_val)
              Sketchup.active_model.select_tool add_edges_tool
              if @skip_edg_proj=="No"
                add_edges_tool.add_proj_edges(view)
              else
                add_edges_tool.skip_proj_edges(view)
              end
            end
            when "Right"
            Sketchup.set_status_text("#{@b_p.typename} is picked.")
            @stick_ray_vec=Geom::Vector3d.new(1,0,0)            
            if @refl_ray_opt=="Specify manually"
              Sketchup.set_status_text("Draw line to specify reflection ray direction.")
              specify_refl_vec=Lss_Specify_Dir_Vec_Tool.new(@mshstick_tool,"reflection ray")
              Sketchup.active_model.select_tool specify_refl_vec
            else
              add_edges_tool=Lss_Add_Edges_Tool.new(@msh_group,@grps_set,@stick_ray_vec,@check_self_int,@stick_type,@refl_ray_vec,@offset_val,@magnify_val)
              Sketchup.active_model.select_tool add_edges_tool
              if @skip_edg_proj=="No"
                add_edges_tool.add_proj_edges(view)
              else
                add_edges_tool.skip_proj_edges(view)
              end
            end
            when "Front"
            Sketchup.set_status_text("#{@b_p.typename} is picked.")
            @stick_ray_vec=Geom::Vector3d.new(0,1,0)            
            if @refl_ray_opt=="Specify manually"
              Sketchup.set_status_text("Draw line to specify reflection ray direction.")
              specify_refl_vec=Lss_Specify_Dir_Vec_Tool.new(@mshstick_tool,"reflection ray")
              Sketchup.active_model.select_tool specify_refl_vec
            else
              add_edges_tool=Lss_Add_Edges_Tool.new(@msh_group,@grps_set,@stick_ray_vec,@check_self_int,@stick_type,@refl_ray_vec,@offset_val,@magnify_val)
              Sketchup.active_model.select_tool add_edges_tool
              if @skip_edg_proj=="No"
                add_edges_tool.add_proj_edges(view)
              else
                add_edges_tool.skip_proj_edges(view)
              end
            end
            when "Back"
            Sketchup.set_status_text("#{@b_p.typename} is picked.")
            @stick_ray_vec=Geom::Vector3d.new(0,-1,0)            
            if @refl_ray_opt=="Specify manually"
              Sketchup.set_status_text("Draw line to specify reflection ray direction.")
              specify_refl_vec=Lss_Specify_Dir_Vec_Tool.new(@mshstick_tool,"reflection ray")
              Sketchup.active_model.select_tool specify_refl_vec
            else
              add_edges_tool=Lss_Add_Edges_Tool.new(@msh_group,@grps_set,@stick_ray_vec,@check_self_int,@stick_type,@refl_ray_vec,@offset_val,@magnify_val)
              Sketchup.active_model.select_tool add_edges_tool
              if @skip_edg_proj=="No"
                add_edges_tool.add_proj_edges(view)
              else
                add_edges_tool.skip_proj_edges(view)
              end
            end
            when "Specify manually"
              Sketchup.set_status_text("#{@b_p.typename} is picked.")
              Sketchup.set_status_text("Draw line from grop towards surface to specify stick ray direction.")
              specify_stick_vec=Lss_Specify_Dir_Vec_Tool.new(@mshstick_tool,"stick ray")
              Sketchup.active_model.select_tool specify_stick_vec
          end
        end
      else
        puts "Nothing is picked"
        Sketchup.set_status_text("Nothing is picked. Click on group (green square bound marks).")
      end
    end
end #class Lss_Msh_Stick_Tool

class Lss_Add_Edges_Tool
    def initialize(msh_group,grps_set,stick_ray_vec,check_self_int,stick_type,refl_ray_vec,offset_val,magnify_val)
      @msh_group=msh_group
      @grps_set=grps_set
      @stick_ray_vec=stick_ray_vec
      @check_self_int=check_self_int
      @stick_type=stick_type

      @refl_ray_vec=refl_ray_vec
      @offset_val=offset_val
      @magnify_val=magnify_val
      
      @entities=Sketchup.active_model.active_entities
      @line_pt1=nil
      @line_pt2=nil
      @line_pt11=nil
      @line_pt22=nil
      @yellow_bnds=nil
      @poly_pnts=nil
      @edg_no=nil
      @edg_no_prev=nil
      @msh_ents_count=0
    end
    
    def reset
      @line_pt1=nil
      @line_pt2=nil
      @line_pt11=nil
      @line_pt22=nil
      @yellow_bnds=nil
      @poly_pnts=nil
      @edg_no=nil
      @edg_no_prev=nil
      @msh_ents_count=0
    end
    
    def activate
      @allow_projection_add=true
      @break_timer=false
    end
    
    def draw(view)
      if @edg_no and @msh_ents_count>0 and @grps_set.length>0
        percents_ready=(100*@edg_no/(@msh_ents_count*@grps_set.length)).round
        txt_pt3d=Geom::Point3d.new(@msh_group.bounds.max)
        txt_pt=view.screen_coords(txt_pt3d)
        progr_char="|"
        rest_char="_"
        progr_string=progr_char*((percents_ready/5).round)+rest_char*(20-(percents_ready/5).round)
        txt_string="#{percents_ready}% #{progr_string}"
        view.draw_text(txt_pt, txt_string)
      end
      if @line_pt11!=@line_pt1 and @line_pt22!=@line_pt2
        status = view.drawing_color="red" 
        view.draw_line @line_pt1,@line_pt2
        @line_pt11=@line_pt1
        @line_pt22=@line_pt2
      end
      if @yellow_bnds
        @yellow_bnds.each{|y_bnd| self.draw_yellow_grp_bnds(y_bnd,view)}
      end
      if @poly_pnts
        status = view.drawing_color="green"
        view.draw_polyline @poly_pnts
      end
    end
    
    def draw_yellow_grp_bnds(bnds,view)
      corner_no=0
      while corner_no<=7
        view.draw_points bnds.corner(corner_no), 6, 2,"yellow"
        corner_no+=1
      end
    end
    
    def deactivate(view)
      view.invalidate
      @line_pt1=nil
      @line_pt2=nil
      @line_pt11=nil
      @line_pt22=nil
      @yellow_bnds=nil
      @poly_pnts=nil
      @msh_ents_count=0
    end
    
    def check_self_intersect(pt,grp)
      model = Sketchup.active_model 
      model_bb = model.bounds
      bounds_offset_len=model_bb.max.distance(model_bb.min)/2.0
      bounds_offset_vec=Geom::Vector3d.new(@stick_ray_vec)
      bounds_offset_vec.length=bounds_offset_len
      grp_center_plane=[Geom::Point3d.new(grp.bounds.center),Geom::Vector3d.new(@stick_ray_vec)] 
      pt_dist=pt.distance_to_plane(grp_center_plane)
      stick_ray=[pt,@stick_ray_vec]
      @msh_group.entities.each{|fc|
        if fc
          if fc.typename=="Face"
            classify1=fc.classify_point(pt)
            break if classify1==1 or classify1==2 or classify1==4
            face_bnds=Geom::BoundingBox.new
            fc.vertices.each{|vrt| face_bnds.add(vrt.position.transform(@msh_group.transformation))}
            face_cntr_min=face_bnds.center.offset(bounds_offset_vec)
            face_cntr_max=face_bnds.center.offset(bounds_offset_vec.reverse)
            face_bnds.add(face_cntr_min,face_cntr_max)
            if face_bnds.contains?(pt)
              fc_tr_pnts=Array.new
              fc.vertices.each{|vrt| fc_tr_pnts+=[vrt.position.transform(@msh_group.transformation)]}
              face_plane=Geom.fit_plane_to_points(fc_tr_pnts)
              int_pt=Geom.intersect_line_plane(stick_ray, face_plane)
              int_classify=fc.classify_point(int_pt.transform(@msh_group.transformation.inverse)) if int_pt
              if int_classify==1 or int_classify==2 #or int_classify==4 #were problems with vertex coincidense
                int_pt_dist=int_pt.distance_to_plane(grp_center_plane)
                if pt_dist>int_pt_dist+0.00001
                  @allow_projection_add=false
                  break
                end
              end
            end
          end
        end
      }
    end
    
    def skip_proj_edges(view)
      @stick_instances=Array.new
      model = Sketchup.active_model 
      @grps_set.each{|init_grp| 
        stick_instance=Lss_Stick_Group.new(@msh_group,@check_self_int,@stick_ray_vec,@stick_type,@refl_ray_vec,@offset_val,@magnify_val)
        @stick_instances+=[stick_instance]
      }
      @stick_instances.each{|st_inst| st_inst.get_stick_instances(@stick_instances)}
      @grps_set.each{|init_grp|
        grp_no=@grps_set.to_a.index(init_grp)
        stick_instance=@stick_instances[grp_no]
        Sketchup.active_model.select_tool stick_instance
        stick_instance.stick_group(view,init_grp.copy)
      }
      status=model.commit_operation
    end
    
    def add_proj_edges(view)
      model = Sketchup.active_model 
      model_bb = model.bounds
      bounds_offset_len=model_bb.max.distance(model_bb.min)
      bounds_offset_vec=Geom::Vector3d.new(@stick_ray_vec)
      bounds_offset_vec.length=bounds_offset_len
      @stick_instances=Array.new
      vert_edg_grp=nil
      model = Sketchup.active_model 
      @yellow_bnds=Array.new
      @msh_ents_count=@msh_group.entities.length
      @grps_set.each{|init_grp| 
        stick_instance=Lss_Stick_Group.new(@msh_group,@check_self_int,@stick_ray_vec,@stick_type,@refl_ray_vec,@offset_val,@magnify_val)
        @stick_instances+=[stick_instance]
      }
      @stick_instances.each{|st_inst| st_inst.get_stick_instances(@stick_instances)}
      @grps_set.each{|init_grp|
        init_grp_bb=init_grp.bounds
        init_grp_center_min=init_grp_bb.center.offset(bounds_offset_vec)
        init_grp_center_max=init_grp_bb.center.offset(bounds_offset_vec.reverse)
        init_grp_bb.add(init_grp_center_min,init_grp_center_max)
        status = model.start_operation "Adding edges to geometry group (#{init_grp})"
        grp=init_grp.copy
        grp_min_plane=[Geom::Point3d.new(grp.bounds.min),Geom::Vector3d.new(@stick_ray_vec)] 
        grp_max_plane=[Geom::Point3d.new(grp.bounds.max),Geom::Vector3d.new(@stick_ray_vec)] 
        grp_bnds=grp.bounds
        @yellow_bnds+=[grp.bounds]
        edg_no=0
        @edg_no=0
        timer_id=UI.start_timer(0.01,true) {
          view.invalidate
          edg=@msh_group.entities[edg_no]
          if edg.nil?
            grp_no=@grps_set.to_a.index(init_grp)
            stick_instance=@stick_instances[grp_no]
            timer_id=UI.stop_timer timer_id
            Sketchup.active_model.select_tool stick_instance
            stick_instance.stick_group(view,grp)
            status = model.commit_operation
          end
          if @break_timer
            timer_id=UI.stop_timer timer_id
            puts "Edges projections addition was cancelled."
            Sketchup.set_status_text("Edges projections addition was cancelled.")
            status = model.commit_operation
            Sketchup.active_model.select_tool nil
            grp.erase! if grp and not grp.deleted?
            vert_edg_grp.erase! if vert_edg_grp and not vert_edg_grp.deleted?
            self.reset
          end
          if edg
            if edg.typename=="Edge"
              edg_strt=edg.start.position; edg_end=edg.end.position
              edg_strt.transform!(@msh_group.transformation); edg_end.transform!(@msh_group.transformation)
              @allow_projection_add=true
              if not init_grp_bb.contains?(edg_strt) and not init_grp_bb.contains?(edg_end)
                @allow_projection_add=false
              else
                if @check_self_int=="Yes"
                  self.check_self_intersect(edg_strt,grp)
                  if @allow_projection_add==false
                    @allow_projection_add=true
                    self.check_self_intersect(edg_end,grp)
                  end
                end
              end
              if @allow_projection_add
                @line_pt1=edg_strt; @line_pt2=edg_end
                edg_bb=Geom::BoundingBox.new
                strt_ray=[Geom::Point3d.new(edg_strt),@stick_ray_vec]
                end_ray=[Geom::Point3d.new(edg_end),@stick_ray_vec]
                pt1=Geom.intersect_line_plane(strt_ray, grp_min_plane)
                pt2=Geom.intersect_line_plane(strt_ray, grp_max_plane)
                pt3=Geom.intersect_line_plane(end_ray, grp_max_plane)
                pt4=Geom.intersect_line_plane(end_ray, grp_min_plane)
                edg_bb.add(pt1,pt2,pt3,pt4)
                @poly_pnts=[pt1,pt2,pt3,pt4,pt1]
                chk_bb=grp_bnds.intersect(edg_bb)
                if not chk_bb.empty?
                  vert_edg_grp=@entities.add_group
                  begin
                    section_fc_points=[pt1,pt2,pt3,pt4]
                    vert_edg_grp.entities.add_face(section_fc_points)
                    if grp and not grp.deleted?
                      grp.entities.intersect_with true, grp.transformation, grp.entities,grp.transformation, true, vert_edg_grp
                    end
                  rescue
                    puts "Can not make projection edge"
                  end
                  vert_edg_grp.erase!
                end
              end
            end
          end
          edg_no+=1
          @edg_no+=1 if @edg_no
        }
      }
    end
    
    def onCancel(reason, view)
      if reason==0
        @break_timer=true
      end
    end
end #class Lss_Add_Edges_Tool

class Lss_Stick_Group
    def initialize(msh_group,check_self_int,stick_ray_vec,stick_type,refl_ray_vec,offset_val,magnify_val)
      @refl_ray_vec=refl_ray_vec
      @offset_val=offset_val
      @magnify_val=magnify_val
      
      @entities=Sketchup.active_model.active_entities
      @msh_group=msh_group
      @check_self_int=check_self_int
      @stick_ray_vec=Geom::Vector3d.new(stick_ray_vec)
      @stick_type=stick_type
      @break_timer=false
    end
    
    def get_stick_instances(stick_instances)
      @stick_instances=stick_instances
    end
    
    def activate
      
    end
    
    def draw(view)
      if @line_pt11!=@line_pt1 and @line_pt22!=@line_pt2
        status = view.drawing_color="red" 
        view.draw_line @line_pt1,@line_pt2
        @line_pt11=@line_pt1
        @line_pt22=@line_pt2
      end
      if @poly_pnts
        status = view.drawing_color="green"
        view.draw_polyline @poly_pnts
      end
    end

    def find_self_intersect(pt)
      grp=@msh_group #switch
      model = Sketchup.active_model 
      model_bb = model.bounds
      bounds_offset_len=model_bb.max.distance(model_bb.min)/2.0
      bounds_offset_vec=Geom::Vector3d.new(@stick_ray_vec)
      bounds_offset_vec.length=bounds_offset_len
      grp_center_plane=[Geom::Point3d.new(grp.bounds.center),Geom::Vector3d.new(@stick_ray_vec)] 
      pt_dist=pt.distance_to_plane(grp_center_plane)
      stick_ray=[pt,@stick_ray_vec]
      int_pt_arr=Array.new
      @init_grp.entities.each{|fc|
        if fc
          if fc.typename=="Face"
            classify1=fc.classify_point(pt)
            break if classify1==1 or classify1==2 or classify1==4
            face_bnds=Geom::BoundingBox.new
            fc.vertices.each{|vrt| face_bnds.add(vrt.position.transform(@init_grp.transformation))}
            face_cntr_min=face_bnds.center.offset(bounds_offset_vec)
            face_cntr_max=face_bnds.center.offset(bounds_offset_vec.reverse)
            face_bnds.add(face_cntr_min,face_cntr_max)
            if face_bnds.contains?(pt)
              fc_tr_pnts=Array.new
              fc.vertices.each{|vrt| fc_tr_pnts+=[vrt.position.transform(@init_grp.transformation)]}
              face_plane=Geom.fit_plane_to_points(fc_tr_pnts)
              @poly_pnts=Array.new(fc_tr_pnts); @poly_pnts+=[@poly_pnts[0]]
              int_pt=Geom.intersect_line_plane(stick_ray, face_plane)
              int_classify=fc.classify_point(int_pt.transform(@init_grp.transformation.inverse)) if int_pt
              if int_classify==1 or int_classify==2 or int_classify==4
                int_pt_arr+=[int_pt]
              end
            end
          end
        end
      }
      if int_pt_arr.length>0
        min_dist=10000000000
        int_pt_arr.each{|pt1|
          int_pt_dist=pt1.distance_to_plane(grp_center_plane)
          if int_pt_dist<min_dist
            min_dist=int_pt_dist
            @int_pt=Geom::Point3d.new(pt1)
          end
        }
        @inside_offset=@int_pt.vector_to(pt)
      else
        @inside_offset=Geom::Vector3d.new
      end
    end

    def measure_offset(mesh_pt)
      model = Sketchup.active_model 
      model_bb = model.bounds
      bounds_offset_len=model_bb.max.distance(model_bb.min)
      bounds_offset_vec=Geom::Vector3d.new(@stick_ray_vec)
      bounds_offset_vec.length=bounds_offset_len
      mesh_pt_tr=mesh_pt.transform(@init_grp.transformation)
      stick_ray=[mesh_pt_tr,Geom::Vector3d.new(@stick_ray_vec)]
      grp_center_plane=[Geom::Point3d.new(@init_grp.bounds.center),@stick_ray_vec] 
      if @stick_type=="Normal"
        stick_vec_ary=[@stick_ray_vec.x.abs,@stick_ray_vec.y.abs,@stick_ray_vec.z.abs]
        max_ind=stick_vec_ary.index(stick_vec_ary.max)
        case max_ind
          when 0
          min_plane_vec=Geom::Vector3d.new(@stick_ray_vec.x,0,0)
          min_plane_vec.normalize!
          when 1
          min_plane_vec=Geom::Vector3d.new(0,@stick_ray_vec.y,0)
          min_plane_vec.normalize!
          when 2
          min_plane_vec=Geom::Vector3d.new(0,0,@stick_ray_vec.z)
          min_plane_vec.normalize!
        end
        grp_min_plane=[Geom::Point3d.new(@init_grp.bounds.min),min_plane_vec]
        curr_pt_min=Geom.intersect_line_plane(stick_ray, grp_min_plane)
        grp_max_plane=[Geom::Point3d.new(@init_grp.bounds.max),min_plane_vec]
        curr_pt_max=Geom.intersect_line_plane(stick_ray, grp_max_plane)
        inside_offset_min=curr_pt_min.vector_to(mesh_pt_tr)
        inside_offset_max=curr_pt_max.vector_to(mesh_pt_tr)
        if inside_offset_min.length>0
          if inside_offset_min.samedirection?(@stick_ray_vec)
            inside_offset=Geom::Vector3d.new(inside_offset_min.reverse)
          else
            inside_offset=Geom::Vector3d.new(inside_offset_min)
          end
        else
          inside_offset=Geom::Vector3d.new(0,0,0)
        end
      else
        self.find_self_intersect(mesh_pt_tr)
        inside_offset=Geom::Vector3d.new(@inside_offset)
      end
      int_pt_arr=Array.new
      fc_arr=Array.new
      pt_classify_arr=Array.new
      @msh_group.entities.each{|fc|
        if fc
          if fc.typename=="Face"
            #~ face_bnds=Geom::BoundingBox.new
            #~ face_cntr_min=face_bnds.center.offset(bounds_offset_vec)
            #~ face_cntr_max=face_bnds.center.offset(bounds_offset_vec.reverse)
            #~ fc.vertices.each{|vrt| face_bnds.add(vrt.position.transform(@msh_group.transformation))}
            #~ face_bnds.add(face_cntr_min,face_cntr_max)
            #~ if face_bnds.contains?(mesh_pt_tr)
              fc_tr_pnts=Array.new
              fc.vertices.each{|vrt| fc_tr_pnts+=[vrt.position.transform(@msh_group.transformation)]}
              face_plane=Geom.fit_plane_to_points(fc_tr_pnts)
              int_pt=Geom.intersect_line_plane(stick_ray, face_plane)
              int_classify=fc.classify_point(int_pt.transform(@msh_group.transformation.inverse)) if int_pt
              if int_classify==1 or int_classify==2 or int_classify==4
                int_pt_arr+=[int_pt]
                pt_classify_arr+=[int_classify]
                case int_classify
                  when 1
                  fc_arr+=[fc]
                  when 2
                  fc.edges.each{|fc_edg|
                    edg_line=fc_edg.line
                    if int_pt.transform(@msh_group.transformation.inverse).distance_to_line(edg_line)==0
                      fc_arr+=[fc_edg.faces]
                    end
                  }
                  when 4
                  fc.vertices.each{|fc_vrt|
                    if int_pt.transform(@msh_group.transformation.inverse).distance(fc_vrt.position)<0.001
                      fc_arr+=[fc_vrt.faces]
                    end
                  }
                end
              end
            #~ end
          end
        end
      }
      min_dist=10000000000
      int_pt_arr.each{|pt|
        int_pt_dist=pt.distance_to_plane(grp_center_plane)
        if int_pt_dist<min_dist 
          min_dist=int_pt_dist
          @int_pt=pt
        end
      }
      outside_offset=mesh_pt_tr.vector_to(@int_pt) if @int_pt
      if inside_offset
        case (@refl_ray_vec.length*1000).round
          when 1000
          inside_offset_len=inside_offset.length
          inside_offset=Geom::Vector3d.new(@refl_ray_vec)
          inside_offset.length=inside_offset_len
          when 1732 #square root of 3
          inside_offset_len=inside_offset.length
          int_pt_ind=int_pt_arr.index(@int_pt)
          int_classify=pt_classify_arr[int_pt_ind]
          case int_classify
            when 1
            inside_offset=fc_arr[int_pt_ind].normal
            inside_offset.length=inside_offset_len
            when 2
            inside_offset=Geom::Vector3d.new
            fc_arr[int_pt_ind].each{|fc| inside_offset+=fc.normal}
            inside_offset.length=inside_offset_len
            when 4
            inside_offset=Geom::Vector3d.new
            fc_arr[int_pt_ind].each{|fc| inside_offset+=fc.normal}
            inside_offset.length=inside_offset_len
          end
        end
      end
      if inside_offset
        if inside_offset.length>0
          inside_offset_len=inside_offset.length
          inside_offset_len_new=inside_offset_len*@magnify_val
          inside_offset.length=inside_offset_len_new
        end
      end
      @offset_vec=Geom::Vector3d.new
      @offset_vec+=inside_offset if inside_offset
      if outside_offset
        if outside_offset.length>0
          outside_offset_len_new=outside_offset.length-@offset_val
          outside_offset.length=outside_offset_len_new
        end
        @offset_vec+=outside_offset
      end
    end
    
    def stick_group(view,grp)
      model = Sketchup.active_model 
      status = model.start_operation "Geometry group (#{@init_grp}) deformation"
      face_mesh=Geom::PolygonMesh.new
      @init_grp=grp
      @init_grp.entities.each{|fc|
        mesh_pnts=Array.new
        if fc.typename=="Face"
          single_face_mesh=fc.mesh
          single_face_mesh.points.each{|pnt|
            mesh_pnts+=[pnt]
          }
          mesh_pnts.uniq!
          mesh_pnts.each{|pnt| face_mesh.add_point(pnt)}
          polygon_ind=1
          while polygon_ind<=single_face_mesh.polygons.length
            face_pnts=single_face_mesh.polygon_points_at(polygon_ind)
            begin
              face_mesh.add_polygon(face_pnts)
            rescue
              puts "Failed to add polygon"
            end
            polygon_ind+=1
          end
        end
      }
      # Progress bar init
      @pb = StickGroupProgressBar.new(face_mesh.points.length)
      pt_ind=1
      timer_id=UI.start_timer(0.01,true) {
        stick_inst=@stick_instances[rand(@stick_instances.length-1)]
        Sketchup.active_model.select_tool stick_inst
        @pb.update(pt_ind)
        mesh_pt=face_mesh.point_at(pt_ind)
        view.invalidate
        self.measure_offset(mesh_pt)
        mesh_pt.offset!(@offset_vec)
        #########################################
        @line_pt1=mesh_pt.transform(@init_grp.transformation)
        @line_pt2=@line_pt1.offset(@offset_vec.reverse) if @line_pt1
        #########################################
        face_mesh.set_point(pt_ind,mesh_pt)
        if pt_ind>=face_mesh.points.length
          timer_id=UI.stop_timer timer_id
          @new_group=Sketchup.active_model.active_entities.add_group
          @new_group.transform! @init_grp.transformation
          @new_group.entities.add_faces_from_mesh(face_mesh,12)
          @init_grp.erase!
          Sketchup.active_model.select_tool nil
          status = model.commit_operation
          puts "The group sticked."
        end
        if @break_timer
          puts "Sticking process was cancelled."
          Sketchup.set_status_text("Sticking process was cancelled.")
          timer_id=UI.stop_timer timer_id
          @init_grp.erase!
          Sketchup.active_model.select_tool nil
          status = model.commit_operation
        end
        pt_ind+=1
      }
    end
    
    def onCancel(reason, view)
      if reason==0
        @break_timer=true
      end
    end
    
    def deactivate(view)
      view.invalidate
      @line_pt1=nil
      @line_pt2=nil
      @line_pt11=nil
      @line_pt22=nil
      @poly_pnts=nil
    end
  end #class Lss_Stick_Group
  
class StickGroupProgressBar 

@@err_total_notnumeric = "ProgressBar: Total must be a positive integer."
@@err_count_notnumeric = "ProgressBar: Iteration Count must be numeric"
@@end_time             = "End @ " 
@@progresschar = ">" ; 
@@initial_block = "-" * 50     # Default progress bar line sequence.

def initialize(total,phase=nil)  
  if (!total.integer? or total < 0)
    raise(ArgumentError,@@err_total_notnumeric) 
    return ; 
    end ; 
  @total = total.to_i ; 
  @phase = phase ; 
  @firsttime = true ; 
  end ; 

def update(iteration) 
  if !iteration.integer? 
    raise(ArgumentError,@@err_count_notnumeric)  
    return ; 
    end ; 
  iteration = [iteration.abs,@total].min  # make sure we don't exceed the total count or have a value < 0
  pct = [1,(iteration*100)/@total].max    # Calculate percentage complete.
                                          # round up to 1% if anything less than 1%.
  end_time = "?" ; 
  if @firsttime then ; 
    # Get the current time of day. 
    # set up an elapsed timer so we can calculate expected end time.  
    @time1 = Time.now ;     # Get current time of day. 
    @firsttime = false ;    # turn off switch 
  else 
    # divide the elapsed time by the pct complete, then multiple that by 100, then add that to the 
    # start time, and that is the expected end time.  
    end_time = Time.at(((((Time.now-@time1).to_f)/pct)*100 + @time1.to_f).to_i).strftime("%H:%M:%S") 
    end ; 
  pct_pos = [pct/2,1].max ;
  current_block = @@initial_block[0,pct_pos-1] << @@progresschar << @@initial_block[pct_pos,@@initial_block.length]
  Sketchup.set_status_text(current_block << "   " << (pct.to_s) << "%. #{@@end_time} " << end_time << " #{@phase}")
  Sketchup.set_status_text(current_block << "   " << (pct.to_s) << "%")
  # if (((Time.now-@time1).to_f/pct)*100) > $cbt_delay
    # $cbt_stop_timer = true
  # end
  end ; 

end ; # class ProgressBar ;


selection=Sketchup.active_model.selection
if( not file_loaded?("mshstick.rb") )
    tool_menu = UI.menu "Plugins" 
    lss_mshstick_tool = Lss_Msh_Stick_Tool.new 
    tool_menu.add_item("Stick selected groups to mesh group") {
      allow_stick=false
      if selection.length>0
        selection.each{|ent|
          if ent.typename=="Group"
            allow_stick=true
          end
        }
      end
      if allow_stick
        lss_mshstick_tool = Lss_Msh_Stick_Tool.new
        lss_mshstick_tool.mshstick_tool=lss_mshstick_tool
        Sketchup.active_model.select_tool lss_mshstick_tool
      else
        UI.messagebox "It is necessary to select some groups of faces before calling the command."
      end
    }
  end
  file_loaded("mshstick.rb")
